module delve.util.vector;

class Vector(T) {
    this() {
        _data.length = 1;
    }

    bool isEmpty() {
        return _length == 0;
    }

    size_t length() {
        return _length;
    }

    void length(size_t l) {
        if (l > _data.length) {
            size_t new_l = _data.length;
            do new_l *= 2; while (new_l < l);
            _data.length = new_l;
        }

        if (l < _length) {
            foreach(t; _data[l.._length]) {
                t = T.init;
            }
        }

        _length = l;
    }

    T first() {
        assert(!isEmpty());
        return _data[0];
    }

    void first(T t) {
        assert(!isEmpty());
        _data[0] = t;
    }

    T last() {
        assert(!isEmpty);
        return _data[_length - 1];
    }

    void last(T t) {
        assert(!isEmpty);
        _data[_length - 1] = t;
    }

    void push(T t) {
        if (_length == _data.length) {
            // Need it to grow
            _data.length = 2 * _data.length;
        }

        _data[_length] = t;
        ++_length;
    }

    T pop() {
        assert(_length > 0);
        --_length;
        auto t = _data[_length];
        _data[_length] = T.init;
        return t;
    }

    void shrink() {
        size_t new_l = _data.length;
        do new_l /= 2; while (new_l > _length * 2);
        _data.length = new_l;
    }

    void capacity(size_t l) {
        if (l > _data.length) {
            size_t new_l = _data.length;
            do new_l *= 2; while (new_l < l);
            _data.length = new_l;
        }
    }

    size_t capacity() {
        return _data.length;
    }

    void clear() {
        while (_length > 0) {
            --_length;
            _data[_length] = T.init;
        }
    }

    int opApply(int delegate (inout T) dg) {
        int result;

        for (size_t i = 0; i < _length; ++i) {
            auto t = _data[i];
            result = dg(t);
            if (result != 0) {
                break;
            }
        }

        return result;
    }

    T opIndex(int i) {
        assert(i < _length && i >= 0);
        return _data[i];
    }

    void opIndex(T t, int i) {
        assert(i < _length && i >= 0);
        _data[i] = t;
    }

    class Iterator {
        T element() {
            return *_element;
        }

        void element(T t) {
            *_element = t;
        }

        void forward() {
            --_element;
        }

        void backward() {
            ++_element;
        }

        void opEquals(Iterator other) {
            return cast(int)(_element == other._element);
        }

        private {
            this(ref T t) {
                _element = &t;
            }

            T * _element;
        }
    }

    /*
    Iterator find(T t) {
    }
    */

    private {
        T[] _data;          // capacity
        size_t _length;
    }
}
unittest {
    auto v = new Vector!(int);

    assert(v.length() == 0);
    assert(v.isEmpty());
    assert(v.capacity() == 1);

    v.push(1);
    assert(v[0] == 1);
    assert(v.length() == 1);
    assert(v.first() == 1);
    assert(v.last() == 1);
    assert(v.capacity() == 1);

    v.push(2);
    assert(v[1] == 2);
    assert(v.length() == 2);
    assert(v.first() == 1);
    assert(v.last() == 2);
    assert(v.capacity() == 2);

    v.pop();
    assert(v.length() == 1);
    assert(v.first() == 1);
    assert(v.last() == 1);
    assert(v.capacity() == 2);

    v.shrink();
    assert(v.capacity() == 1);

    v.first(100);
    assert(v.first == 100);
}
